import itertools
import random
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pytomlpp
import tifffile as tf
from tqdm import tqdm
from cenfind.core.log import get_logger

logger = get_logger(__name__)


@dataclass
class Field:
    name: str
    dataset: "Dataset"

    @property
    def stack(self) -> np.ndarray:
        path_file = str(self.dataset.raw / f"{self.name}{self.dataset.image_type}")

        try:
            data = tf.imread(path_file)
        except FileNotFoundError:
            logger.error(f"File not found (%s)" % path_file)
            raise

        axes_order = self._axes_order()
        if axes_order == "ZCYX":
            data = np.swapaxes(data, 0, 1)
        return data

    @property
    def projection(self) -> np.ndarray:
        _projection_name = f"{self.name}{self.dataset.projection_suffix}.tif"

        path_projection = self.dataset.projections / _projection_name
        try:
            res = tf.imread(str(path_projection))
            return res
        except FileNotFoundError:
            logger.error("File not found (%s). Check the projection suffix..." % path_projection)
            raise

    def channel(self, channel: int) -> np.ndarray:
        return self.projection[channel, :, :]

    def annotation(self, channel) -> np.ndarray:
        """
        Load annotation file from text file given channel
        loaded as row col, row major.
        ! the text format is x, y; origin at top left;
        :param channel:
        :return:
        """
        name = f"{self.name}{self.dataset.projection_suffix}_C{channel}"
        path_annotation = (
                self.dataset.path / "annotations" / "centrioles" / f"{name}.txt"
        )
        try:
            annotation = np.loadtxt(str(path_annotation), dtype=int, delimiter=",")
            if len(annotation) == 0:
                return annotation
            else:
                return annotation[:, [1, 0]]
        except OSError:
            logger.error(f"No annotation found for %s" % path_annotation, exc_info=True)
            raise

    def mask(self, channel) -> np.ndarray:
        mask_name = f"{self.name}{self.dataset.projection_suffix}_C{channel}.tif"
        path_annotation = self.dataset.path / "annotations" / "cells" / mask_name
        if path_annotation.exists():
            return tf.imread(str(path_annotation))
        else:
            raise FileNotFoundError(path_annotation)

    def _axes_order(self) -> str:
        """
        Return a string of the form 'ZYCX' or 'CZYX'
        :return:
        """
        path_raw = str(
            self.dataset.path / "raw" / f"{self.name}{self.dataset.image_type}"
        )
        with tf.TiffFile(path_raw) as tif:
            try:
                order = tif.series[0].axes
            except ValueError(
                    f"Could not retrieve metadata for axes order for {path_raw}"
            ):
                order = None

        return order


@dataclass
class Dataset:
    """
    Represent a dataset structure
    """

    path: Union[str, Path]
    projection_suffix: str = None
    image_type: str = ".ome.tif"
    pixel_size: float = 0.1025
    has_projections: bool = False
    is_setup: bool = False

    def __post_init__(self):
        self.path = Path(self.path)

        if not self.path.is_dir():
            logger.error(f"Dataset does not exist ({self.path})")
            raise FileNotFoundError

        self.raw = self.path / "raw"
        self.projections = self.path / "projections"
        self.predictions = self.path / "predictions"
        self.visualisation = self.path / "visualisations"
        self.statistics = self.path / "statistics"
        self.vignettes = self.path / "vignettes"
        self.logs = self.path / "logs"
        self.path_annotations = self.path / "annotations"
        self.path_annotations_centrioles = self.path_annotations / "centrioles"
        self.path_annotations_cells = self.path_annotations / "cells"

        if self.projection_suffix is None:
            try:
                metadata = pytomlpp.load(self.path / "metadata.toml")
                self.projection_suffix = metadata["projection_suffix"]
            except FileNotFoundError:
                logger.error("Metadata file not found (%s)" % str(self.path / "metadata.toml"))
                raise

    def setup(self) -> None:
        """
        Create folders for projections, predictions, statistics, visualisation and vignettes.
        Collect field names into fields.txt
        """
        self.projections.mkdir(exist_ok=True)
        self.predictions.mkdir(exist_ok=True)
        (self.predictions / "centrioles").mkdir(exist_ok=True)
        (self.predictions / "nuclei").mkdir(exist_ok=True)
        self.statistics.mkdir(exist_ok=True)
        self.visualisation.mkdir(exist_ok=True)
        self.vignettes.mkdir(exist_ok=True)
        self.logs.mkdir(exist_ok=True)
        self.path_annotations.mkdir(parents=True, exist_ok=True)
        self.path_annotations_centrioles.mkdir(exist_ok=True)
        self.path_annotations_cells.mkdir(exist_ok=True)

        self.has_projections = bool(len([f for f in self.projections.iterdir()]))
        self.is_setup = True

    @property
    def fields(self) -> List[Field]:
        """
        Construct a list of Fields using the fields listed in fields.txt.
        """
        fields_path = self.path / "fields.txt"
        with open(fields_path, "r") as f:
            fields_list = f.read().splitlines()
        return [Field(field_name, self) for field_name in fields_list]

    def write_fields(self) -> None:
        """
        Write field names to fields.txt.
        """
        if not self.is_setup:
            self.setup()

        if self.has_projections:
            folder = self.projections
        else:
            folder = self.raw

        def _field_name(file_name: str):
            return file_name.split(".")[0].rstrip(self.projection_suffix)

        fields = [
            _field_name(str(f.name))
            for f in folder.iterdir()
            if not str(f).startswith(".")
        ]

        with open(self.path / "fields.txt", "w") as f:
            for field in fields:
                f.write(field + "\n")

    def write_projections(self, axis=1) -> None:
        for field in tqdm(self.fields):
            projection = field.stack.max(axis)
            tf.imwrite(
                self.projections / f"{field.name}{self.projection_suffix}.tif",
                projection,
                photometric="minisblack",
                imagej=True,
                resolution=(1 / self.pixel_size, 1 / self.pixel_size),
                metadata={"unit": "um"},
            )
        self.has_projections = True

    def split_pairs(self, p=0.9, seed=1993) -> tuple[list[Field], list[Field]]:
        """
        Split a list of pairs (field, channel).

        :param fields
        :param p the train proportion, default to .9
        :param seed the seed to reproduce the splitting
        :return train_split, test_split
        """

        random.seed(seed)
        size = len(self.fields)
        split_idx = int(p * size)
        shuffled = random.sample(self.fields, k=size)
        split_test = shuffled[split_idx:]
        split_train = shuffled[:split_idx]

        return split_train, split_test

    def pairs(
            self, split: str = None, channel_id: int = None
    ) -> List[Tuple["Field", int]]:
        """
        Fetch the fields of view for train or test
        :param channel_id:
        :param split: all, test or train
        :return: a list of tuples (fov name, channel id)
        """

        if split is None:
            return self.read_split("train", channel_id) + self.read_split(
                "test", channel_id
            )
        else:
            return self.read_split(split, channel_id)

    def read_split(self, split_type, channel_id=None) -> List[Tuple[Field, int]]:
        with open(self.path / f"{split_type}.txt", "r") as f:
            files = f.read().splitlines()

        files = [f.split(",") for f in files if f]
        if channel_id:
            return [(Field(str(f[0]), self), int(channel_id)) for f in files]
        else:
            return [(Field(str(f[0]), self), int(f[1])) for f in files]


def choose_channel(fields: list[Field], channels: list[int]) -> list[tuple[Field, int]]:
    """Pick a channel for each field."""
    return [
        (field, int(channel)) for field, channel in itertools.product(fields, channels)
    ]



import logging
from pathlib import Path
from types import SimpleNamespace
from typing import List

import cv2
import numpy as np
import pandas as pd
from ortools.linear_solver import pywraplp
from spotipy.utils import points_matching

from cenfind.core.data import Dataset, Field
from cenfind.core.detectors import extract_foci
from cenfind.core.outline import Centre, Contour
from cenfind.core.log import get_logger

logger = get_logger(__name__)


def signed_distance(focus: Centre, nucleus: Contour) -> float:
    """Wrapper for the opencv PolygonTest"""

    result = cv2.pointPolygonTest(nucleus.contour, focus.to_cv2(), measureDist=True)
    return result


def full_in_field(nucleus: Contour, image_shape, fraction) -> bool:
    h, w = image_shape
    pad_lower = int(fraction * h)
    pad_upper = h - pad_lower
    centroid = nucleus.centre.to_numpy()
    if all([pad_lower < c < pad_upper for c in centroid]):
        return True
    return False


def flag(is_full: bool) -> tuple:
    return (0, 255, 0) if is_full else (0, 0, 255)


def assign(
    nuclei: List[Contour], centrioles: List[Centre], vicinity=0
) -> List[Contour]:
    _nuclei = nuclei.copy()
    _centrioles = centrioles.copy()

    num_nuclei = len(_nuclei)
    num_centrioles = len(_centrioles)

    costs = {}

    for i in range(num_nuclei):
        for j in range(num_centrioles):
            dist = signed_distance(_centrioles[j], _nuclei[i])
            costs[i, j] = dist + vicinity
    solver = pywraplp.Solver.CreateSolver("SCIP")

    x = {}
    for i in range(num_nuclei):
        for j in range(num_centrioles):
            x[i, j] = solver.IntVar(0, 1, "")

    for j in range(num_centrioles):
        solver.Add(solver.Sum([x[i, j] for i in range(num_nuclei)]) <= 1)

    # Objective
    objective_terms = []
    for i in range(num_nuclei):
        for j in range(num_centrioles):
            objective_terms.append(costs[i, j] * x[i, j])
    solver.Maximize(solver.Sum(objective_terms))

    # Solve
    status = solver.Solve()
    if status != pywraplp.Solver.OPTIMAL and status != pywraplp.Solver.FEASIBLE:
        raise ValueError("No solution found.")

    for i in range(num_nuclei):
        for j in range(num_centrioles):
            if x[i, j].solution_value() > 0.5:
                logger.debug("Adding Centre %s to Nucleus %s" %(j, i))
                _nuclei[i].add_centrioles(_centrioles[j])

    return _nuclei


# TODO: refactor as a field method
def score(
    field,
    nuclei_scored,
    channel: int,
) -> List[dict]:
    """
    1. Detect foci in the given channels
    2. Detect nuclei
    3. Assign foci to nuclei
    :param channel:
    :param nuclei_channel:
    :param model_foci:
    :param model_nuclei:
    :param field:
    :return: list(foci, nuclei, assigned, scores)
    """
    image_shape = field.projection.shape[1:]
    scores = []
    for nucleus in nuclei_scored:
        scores.append(
            {
                "fov": field.name,
                "channel": channel,
                "nucleus": nucleus.centre.to_numpy(),
                "score": len(nucleus.centrioles),
                "is_full": full_in_field(nucleus, image_shape, 0.05),
            }
        )
    return scores


def field_score_frequency(df, by="field"):
    """
    Count the absolute frequency of number of centriole per well or per field
    :param df: Df containing the number of centriole per nuclei
    :return: Df with absolut frequencies.
    """
    cuts = [0, 1, 2, 3, 4, 5, np.inf]
    labels = "0 1 2 3 4 +".split(" ")

    df = df.set_index(["fov", "channel"])
    result = pd.cut(df["score"], cuts, right=False, labels=labels, include_lowest=True)
    result = result.groupby(["fov", "channel"]).value_counts()
    result.name = "freq_abs"
    result = result.sort_index().reset_index()
    result = result.rename({"score": "score_cat"}, axis=1)
    if by == "well":
        result[["well", "field"]] = result["fov"].str.split("_", expand=True)
        print(result.columns)
        result = result.groupby(["well", "channel", "score_cat"])[["freq_abs"]].sum()
        result = result.reset_index()
        result = result.pivot(index=["well", "channel"], columns="score_cat")
        result.reset_index().sort_values(["channel", "well"])
    else:
        result = result.groupby(["fov", "channel", "score_cat"]).sum()
        result = result.reset_index()
        result = result.pivot(index=["fov", "channel"], columns="score_cat")
        result.reset_index().sort_values(["channel", "fov"])

    return result


def save_foci(foci_list: list[Centre], dst: str, logger=None) -> None:
    if len(foci_list) == 0:
        array = np.array([])
        if logger is not None:
            logger.info("No centriole detected")
        else:
            print("No centriole detected")
    else:
        array = np.asarray(np.stack([c.to_numpy() for c in foci_list]))
        array = array[:, [1, 0]]
    np.savetxt(dst, array, delimiter=",", fmt="%u")


# TODO: refactor as a field method
def field_metrics(
    field: Field,
    channel: int,
    annotation: np.ndarray,
    predictions: np.ndarray,
    tolerance: int,
    threshold: float,
) -> dict:
    """
    Compute the accuracy of the prediction on one field.
    :param threshold:
    :param field:
    :param channel:
    :param annotation:
    :param predictions:
    :param tolerance:
    :return: dictionary of fields
    """
    if all((len(predictions), len(annotation))) > 0:
        res = points_matching(annotation, predictions, cutoff_distance=tolerance)
    else:
        logging.warning(
            "threshold: %f; detected: %d; annotated: %d... Set precision and accuracy to zero"
            % (threshold, len(predictions), len(annotation))
        )
        res = SimpleNamespace()
        res.precision = 0.0
        res.recall = 0.0
        res.f1 = 0.0
        res.tp = (0,)
        res.fp = (0,)
        res.fn = 0
    perf = {
        "dataset": field.dataset.path.name,
        "field": field.name,
        "channel": channel,
        "n_actual": len(annotation),
        "n_preds": len(predictions),
        "threshold": threshold,
        "tolerance": tolerance,
        "tp": res.tp[0] if type(res.tp) == tuple else res.tp,
        "fp": res.fp[0] if type(res.fp) == tuple else res.fp,
        "fn": res.fn[0] if type(res.fn) == tuple else res.fn,
        "precision": np.round(res.precision, 3),
        "recall": np.round(res.recall, 3),
        "f1": np.round(res.f1, 3),
    }
    return perf


# TODO: refactor as a field method
def dataset_metrics(
    dataset: Dataset, split: str, model: Path, tolerance, threshold
) -> list[dict]:
    if type(tolerance) == int:
        tolerance = [tolerance]
    perfs = []
    for field, channel in dataset.pairs(split):
        annotation = field.annotation(channel)
        predictions = extract_foci(field, model, channel, prob_threshold=threshold)
        for tol in tolerance:
            perf = field_metrics(
                field, channel, annotation, predictions, tol, threshold=threshold
            )
            perfs.append(perf)
    return perfs



from numpy.random import seed


import tensorflow as tf

import contextlib
import json
import os
from datetime import datetime
from pathlib import Path

from tqdm import tqdm
import numpy as np
import albumentations as alb

from spotipy.model import SpotNet, Config
from spotipy.utils import points_to_prob, normalize_fast2d

from cenfind.core.data import Dataset
from cenfind.core.config import config_multiscale, transforms

seed(1)
tf.random.set_seed(2)


def read_config(path):
    """
    Read config json file.
    :param path:
    :return:
    """
    with open(path, "r") as config:
        config_dict = json.load(config)
    return Config(**config_dict)


# TODO: Include in Dataset or in torch.Dataset
def load_pairs(
    dataset: Dataset, split: str, sigma: float = 1.5, transform: alb.Compose = None
):
    """
    Load two arrays, the images and the foci masks
    path: the path to the ds
    split: either train or test
    """

    channels = []
    masks = []

    for field, channel in dataset.pairs(split):
        print(field.name, channel)
        data = field.channel(channel)
        foci = field.annotation(channel)

        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            image = normalize_fast2d(data)

        if len(foci) == 0:
            mask = np.zeros(image.shape, dtype="uint16")
        else:
            mask = points_to_prob(
                foci[:, [1, 0]], shape=image.shape, sigma=sigma
            )  # because it works with x, y

        if transform is not None:
            transformed = transform(image=image, mask=mask)
            image = transformed["image"]
            mask = transformed["mask"]

        channels.append(image)
        masks.append(mask)

    return np.stack(channels), np.stack(masks)


def fetch_all_fields(datasets: list[Dataset]):
    all_train_x = []
    all_train_y = []

    all_test_x = []
    all_test_y = []

    for ds in tqdm(datasets):
        train_x, train_y = load_pairs(ds, split="train", transform=transforms)
        test_x, test_y = load_pairs(ds, split="test")
        all_train_x.append(train_x)
        all_train_y.append(train_y)
        all_test_x.append(test_x)
        all_test_y.append(test_y)

    all_train_x = np.concatenate(all_train_x, axis=0)
    all_train_y = np.concatenate(all_train_y, axis=0)

    all_test_x = np.concatenate(all_test_x, axis=0)
    all_test_y = np.concatenate(all_test_y, axis=0)

    return all_train_x, all_train_y, all_test_x, all_test_y


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "train", help="Train a Spotnet model using the datasets"
    )
    parser.add_argument("datasets", type=Path, nargs="+", help="Path to the dataset")
    parser.add_argument(
        "--model_path", type=Path, required=True, help="Path to the model fit"
    )
    parser.add_argument("--epochs", type=int, required=True, help="Number of epochs")

    return parser


def run(args):
    datasets = [Dataset(path) for path in args.datasets]

    all_train_x, all_train_y, all_test_x, all_test_y = fetch_all_fields(datasets)

    time_stamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    model_multiscale = SpotNet(
        config_multiscale, name=time_stamp, basedir=args.model_path
    )

    model_multiscale.train(
        all_train_x,
        all_train_y,
        validation_data=(all_test_x, all_test_y),
        epochs=args.epochs,
    )



from pathlib import Path

from dotenv import dotenv_values
from labelbox import Client

from cenfind.core.data import Dataset


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "upload", help="Upload the vignettes to labelbox"
    )
    parser.add_argument(
        "path", type=Path, help="Path to the dataset with existing vignettes"
    )
    parser.add_argument("--env", type=Path, help="Path to the .env file")

    return parser


def run(args):
    ds = Dataset(args.path)

    config = dotenv_values(args.env)
    client = Client(api_key=config["LABELBOX_API_KEY"])

    dataset = client.create_dataset(name=f"{ds.path.name}", iam_integration=None)

    asset = [
        {"row_data": str(path), "external_id": path.name}
        for path in sorted((ds.path / "vignettes").iterdir())
    ]
    dataset.create_data_rows(asset)



from spotipy.utils import normalize_fast2d
from cenfind.core.detectors import get_model
from pathlib import Path
from matplotlib import pyplot as plt
from matplotlib import patches
from matplotlib_scalebar.scalebar import ScaleBar
import cv2
import numpy as np
from skimage import exposure
from cenfind.core.outline import to_8bit
from cenfind.core.data import Dataset, Field

plt.rcParams["font.family"] = "Helvetica"


def main():
    model = get_model('models/dev/5785b6d9-f09b-4486-af65-0a923c8ae533')

    dataset = Dataset(Path('/Users/buergy/Dropbox/epfl/datasets/RPE1wt_CEP152+GTU88+PCNT_1'))
    field = Field('RPE1wt_CEP152+GTU88+PCNT_1_MMStack_1-Pos_001_002', dataset)
    data = field.projection[2, :, :]
    dna = field.projection[0, :, :]

    percentile = .2
    width = 32
    x, y = 1214, 406
    top_left = slice(x, x + width)
    bottom_right = slice(y, y + width)

    composite = np.zeros_like(data)
    composite = cv2.cvtColor(composite, cv2.COLOR_GRAY2RGB)
    composite[:, :, 2] = to_8bit(dna)
    composite[:, :, 1] = to_8bit(data)
    percentiles = np.percentile(composite, (percentile, 100 - percentile))
    scaled = exposure.rescale_intensity(composite,
                                        in_range=tuple(percentiles),
                                        out_range='uint8')
    cell = composite[top_left, bottom_right]

    x = normalize_fast2d(data)
    prob_thresh = .5

    foci = model.predict(x,
                         prob_thresh=prob_thresh,
                         show_tile_progress=False)

    fig, ax = plt.subplots(1, 3, figsize=(7.2, 7.2 / 3))

    ax[0].imshow(scaled, interpolation='nearest')
    roi = patches.Rectangle((406, 1214), width, width, linewidth=.5, edgecolor='white', facecolor='none')
    ax[0].add_patch(roi)

    scalebar_fov = ScaleBar(dx=.1025,
                            units='um',
                            location='lower right',
                            box_alpha=0, color='white')
    ax[0].add_artist(scalebar_fov)
    ax[0].set_title('DNA + Centrin')

    ax[1].imshow(cell, interpolation='nearest')
    scalebar_fov = ScaleBar(dx=.1025, units='um',
                            location='lower right',
                            box_alpha=0, color='white')
    ax[1].add_artist(scalebar_fov)
    ax[1].set_title('Centrin')

    ax[2].imshow(foci[0][top_left, bottom_right], vmin=0, vmax=1,
                 cmap='cividis', interpolation='nearest')
    ax[2].set_title('Prediction')

    for a in ax:
        a.axis('off')
    fig.savefig('out/test.png', bbox_inches='tight', dpi=300)

    return 0


if __name__ == '__main__':
    main()



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import cv2
import numpy as np

# import numpy.ma as ma
from skimage.draw import disk
from skimage.exposure import rescale_intensity

from cenfind.core.data import Field


@dataclass(eq=True, frozen=False)
class ROI(ABC):
    """Abstract class to represent any region of interest"""

    @property
    @abstractmethod
    def centre(self):
        pass

    @abstractmethod
    def draw(self, plane, color, marker_type, marker_size):
        pass


@dataclass(eq=True, frozen=False)
class Centre(ROI):
    position: tuple
    idx: int = 0
    label: str = ""
    confidence: float = 0
    parent: "Centre" = None

    @property
    def centre(self):
        row, col = self.position
        return int(row), int(col)

    # def intensity(self, field: Field, channel: int):
    #     data = field.channel(channel)
    #     result = data[self.position]
    #     return result

    def draw(
        self,
        image,
        color=(0, 255, 0),
        annotation=True,
        marker_type=cv2.MARKER_SQUARE,
        marker_size=8,
    ):
        r, c = self.centre
        offset_col = int(0.01 * image.shape[1])

        if annotation:
            cv2.putText(
                image,
                f"{self.label} {self.idx}",
                org=(c + offset_col, r),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.4,
                thickness=1,
                color=color,
            )

        return cv2.drawMarker(
            image, (c, r), color, markerType=marker_type, markerSize=marker_size
        )

    def to_numpy(self):
        return np.asarray(self.centre)

    def to_cv2(self):
        y, x = self.centre
        return x, y


@dataclass(eq=True, frozen=False)
class Contour(ROI):
    """Represent a blob using the row-column scheme."""

    contour: np.ndarray
    label: str = ""
    idx: int = 0
    confidence: float = 0
    centrioles: list = field(default_factory=list)

    @property
    def centre(self):
        moments = cv2.moments(self.contour)

        centre_x = int(moments["m10"] / (moments["m00"] + 1e-5))
        centre_y = int(moments["m01"] / (moments["m00"] + 1e-5))
        return Centre((centre_y, centre_x), self.idx, self.label, self.confidence)

    def draw(self, image, color=(0, 255, 0), annotation=True, thickness=2, **kwargs):
        r, c = self.centre.centre
        cv2.drawContours(image, [self.contour], -1, color, thickness=thickness)
        if annotation:
            cv2.putText(
                image,
                f"{self.label}{self.idx}",
                org=(c, r),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.8,
                thickness=2,
                color=color,
            )
            cv2.drawMarker(
                image, (c, r), color, markerType=cv2.MARKER_STAR, markerSize=10
            )
        return image

    def add_centrioles(self, centriole: Centre):
        self.centrioles.append(centriole)
        return 0

    # def intensity(contour, field: Field, channel: int):
    #     data = field.channel(channel)
    #     label = np.zeros_like(data)
    #     cv2.drawContours(label, [contour], -1, 1, thickness=-1)
    #     result = (data * ma.masked_not_equal(label, 1)).sum()
    #     return result


def resize_image(data, factor=256):
    height, width = data.shape
    shrinkage_factor = int(height // factor)
    height_scaled = int(height // shrinkage_factor)
    width_scaled = int(width // shrinkage_factor)
    data_resized = cv2.resize(
        data,
        dsize=(height_scaled, width_scaled),
        fx=1,
        fy=1,
        interpolation=cv2.INTER_NEAREST,
    )
    return data_resized


def draw_foci(data: np.ndarray, foci: list[Centre], radius=2) -> np.ndarray:
    mask = np.zeros(data.shape, dtype="uint8")
    for f in foci:
        r, c = f.to_numpy()
        rr, cc = disk((r, c), radius)
        try:
            mask[rr, cc] = 250
        except IndexError:
            continue
    return mask


def _color_channel(data, color, out_range):
    """
    Create a colored version of a channel image
    :return:
    """
    data = rescale_intensity(data, out_range=out_range)
    b = np.multiply(data, color[0], casting="unsafe")
    g = np.multiply(data, color[1], casting="unsafe")
    r = np.multiply(data, color[2], casting="unsafe")
    res = cv2.merge([b, g, r])
    return res


def create_vignette(field: Field, marker_index: int, nuclei_index: int):
    """
    Normalise all markers
    Represent them as blue
    Highlight the channel in green
    :param field:
    :param nuclei_index:
    :param marker_index:
    :return:
    """
    layer_nuclei = field.channel(nuclei_index)
    layer_marker = field.channel(marker_index)

    nuclei = _color_channel(layer_nuclei, (1, 0, 0), "uint8")
    marker = _color_channel(layer_marker, (0, 1, 0), "uint8")

    res = cv2.addWeighted(marker, 1, nuclei, 0.5, 0)
    res = cv2.putText(
        res,
        f"{field.name} {marker_index}",
        (100, 100),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.8,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )

    return res


def visualisation(
    field: Field,
    nuclei: list,
    centrioles: list,
    channel_centrioles: int,
    channel_nuclei: int,
) -> np.ndarray:
    background = create_vignette(
        field, marker_index=channel_centrioles, nuclei_index=channel_nuclei
    )

    if nuclei is None:
        return background

    for nucleus in nuclei:
        background = nucleus.draw(background, annotation=False)
        background = nucleus.centre.draw(background, annotation=False)
        for centriole in centrioles:
            background = centriole.draw(background, annotation=False)

        for centriole in nucleus.centrioles:
            cv2.arrowedLine(
                background,
                centriole.to_cv2(),
                nucleus.centre.to_cv2(),
                color=(0, 255, 0),
                thickness=1,
            )

    return background



import sys
from pathlib import Path

import pandas as pd

from cenfind.core.constants import UNITS
from cenfind.core.data import Dataset
from cenfind.core.measure import field_score_frequency


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "analyse",
        help="Analyse the scoring and compute summary table and visualisation",
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset")
    parser.add_argument(
        "--by", type=str, required=True, help="Grouping (field or well)"
    )

    return parser


def run(args):
    dataset = Dataset(args.dataset)

    if args.by not in UNITS:
        print(f"ERROR: `by` is not in {UNITS} ({{args.by}})")
        sys.exit(2)

    scores = pd.read_csv(dataset.statistics / "scores_df.tsv", sep="\t")

    try:
        binned = field_score_frequency(scores, by=args.by)
    except ValueError as e:
        print(
            "The field value of `fov` does not conform with the syntax `<WELL>_<FOVID>`(%s) (%s)"
            % (scores.loc[0, "fov"], e)
        )
        sys.exit()

    path_stats = dataset.statistics / "statistics.tsv"
    binned.to_csv(path_stats, sep="\t", index=True)
    print(f"Writing statistics to {path_stats}")

    if args.by != "well":
        return 0

    # data = pd.read_csv(path_stats , sep='\t', header=[0, 1, 2])
    # vmin = 'A1'
    # vmax = 'A4'
    # data = prepare_data(data)
    # summed = reduce_data(data)
    # fig = generate_figure(summed, vmin=vmin, vmax=vmax)
    # fig.savefig(dataset.statistics / 'layout_score.png', dpi=300)



import sys
import os
import contextlib
from pathlib import Path

import pandas as pd
import tifffile as tif

from cenfind.core.data import Dataset
from cenfind.core.outline import visualisation


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "evaluate", help="Evaluate the model on the test split of the dataset"
    )

    parser.add_argument("dataset", type=Path, help="Path to the dataset folder")
    parser.add_argument("model", type=Path, help="Path to the model")
    parser.add_argument(
        "--performances_file",
        type=Path,
        help="Path of the performance file, STDOUT if not specified",
    )
    parser.add_argument(
        "--tolerance",
        type=int,
        nargs="+",
        default=3,
        help="Distance in pixels below which two points are deemed matching",
    )
    parser.add_argument(
        "--channel_nuclei",
        type=int,
        required=True,
        help="Channel index for nuclei segmentation, e.g., 0 or 3",
    )
    parser.add_argument(
        "--channel_centrioles",
        nargs="+",
        type=int,
        required=True,
        help="Channel indices to analyse, e.g., 1 2 3",
    )
    parser.add_argument(
        "--vicinity",
        type=int,
        default=-5,
        help="Distance threshold in micrometer (default: -5 um)",
    )

    return parser


def run(args):
    dataset = Dataset(args.dataset)
    if not any(dataset.path_annotations_centrioles.iterdir()):
        print(
            f"ERROR: The dataset {dataset.path.name} has no annotation. You can run `cenfind predict` instead"
        )
        sys.exit(2)

    from cenfind.core.measure import dataset_metrics, field_score

    _, performance = dataset_metrics(
        dataset, split="test", model=args.model, tolerance=args.tolerance, threshold=0.5
    )
    from stardist.models import StarDist2D

    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        model_stardist = StarDist2D.from_pretrained("2D_versatile_fluo")

    path_visualisation_model = dataset.visualisation / args.model.name
    path_visualisation_model.mkdir(exist_ok=True)

    for field, channel in dataset.pairs("test"):
        foci, nuclei, assigned, score = field_score(
            field=field,
            model_nuclei=model_stardist,
            model_foci=args.model,
            nuclei_channel=args.channel_nuclei,
            vicinity=args.vicinity,
            channel=channel,
        )
        vis = visualisation(
            field=field,
            foci=foci,
            channel_centrioles=channel,
            nuclei=nuclei,
            channel_nuclei=args.channel_nuclei,
            assigned=assigned,
        )
        tif.imwrite(path_visualisation_model / f"{field.name}_C{channel}_pred.png", vis)

    performance_df = pd.DataFrame(performance)
    performance_df = performance_df.set_index("field")
    if args.performances_file:
        performance_df.to_csv(args.performances_file)
        print(performance_df)
        print("Performances have been saved under %s" % args.performances_file)
    else:
        print(performance_df)
        print("Performances are ONLY displayed, not saved")



import contextlib
import functools
import logging
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
from pathlib import Path
from typing import List

import cv2
import numpy as np
import tensorflow as tf
from csbdeep.utils import normalize
from skimage.measure import label, regionprops
from spotipy.model import SpotNet
from spotipy.utils import normalize_fast2d
from stardist.models import StarDist2D

from cenfind.core.data import Field
from cenfind.core.outline import Centre, Contour, draw_foci, resize_image

np.random.seed(1)
tf.random.set_seed(2)
tf.get_logger().setLevel(logging.ERROR)

@functools.lru_cache(maxsize=None)
def get_model(model):
    path = Path(model)
    if not path.is_dir():
        raise (FileNotFoundError(f"{path} is not a directory"))

    return SpotNet(None, name=path.name, basedir=str(path.parent))


def extract_foci(
    data: Field,
    foci_model_file: Path,
    channel: int,
    prob_threshold=0.5,
    min_distance=2,
    **kwargs,
) -> List[Centre]:
    """
    Detect centrioles as row, col, row major
    :param data:
    :param foci_model_file:
    :param channel:
    :param prob_threshold:
    :param min_distance:
    :param kwargs:
    :return:
    """
    data = data.channel(channel)
    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        data = normalize_fast2d(data)
        model = get_model(foci_model_file)
        _, points_preds = model.predict(
            data, prob_thresh=prob_threshold, min_distance=min_distance, verbose=False
        )
    foci = [
        Centre((r, c), f_id, "Centriole") for f_id, (r, c) in enumerate(points_preds)
    ]

    centrosomes_mask = np.zeros(data.shape, dtype="uint8")
    centrosomes_mask = draw_foci(centrosomes_mask, foci, radius=min_distance * 2)

    centrosomes_map = label(centrosomes_mask)
    centrosomes_centroids = regionprops(centrosomes_map)

    for f in foci:
        foci_index = centrosomes_map[f.centre]
        centrosome_centroid = centrosomes_centroids[foci_index - 1].centroid
        centrosome_centroid = tuple(int(c) for c in centrosome_centroid)
        f.parent = Centre(centrosome_centroid, label="Centrosome")

    return foci


def extract_nuclei(
    field: Field, channel: int, factor: int, model: StarDist2D = None, annotation=None
) -> List[Contour]:
    """
    Extract the nuclei from the nuclei image
    :param field:
    :param channel:
    :param factor: the factor related to pixel size
    :param model:
    :param annotation: a mask with pixels labelled for each centre

    :return: List of Contours.

    """
    if model is None:
        from stardist.models import StarDist2D
        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            model = StarDist2D.from_pretrained("2D_versatile_fluo")

    if annotation is not None:
        labels = annotation
    elif model is not None:
        data = field.channel(channel)
        data_resized = resize_image(data, factor)
        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            labels, _ = model.predict_instances(normalize(data_resized))
        labels = cv2.resize(
            labels, dsize=data.shape, fx=1, fy=1, interpolation=cv2.INTER_NEAREST
        )

    else:
        raise ValueError("Please provide either an annotation or a model")

    labels_id = np.unique(labels)

    cnts = []
    for nucleus_id in labels_id:
        if nucleus_id == 0:
            continue
        sub_mask = np.zeros_like(labels, dtype="uint8")
        sub_mask[labels == nucleus_id] = 1
        contour, _ = cv2.findContours(
            sub_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )

        cnts.append(contour[0])

    contours = tuple(cnts)
    contours = [
        Contour(c, "Nucleus", c_id, confidence=-1) for c_id, c in enumerate(contours)
    ]

    return contours



import argparse
import sys
from pathlib import Path

from cenfind.core.data import Dataset
from cenfind.core.log import get_logger


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("squash", help="Write z-projections.")
    parser.add_argument("dataset", type=Path, help="Path to the dataset folder")

    return parser


def run(args):
    """
    Squash the raw fields of view.
    """

    dataset = Dataset(args.dataset)
    logger = get_logger(__name__, console=1, file=dataset.logs / 'cenfind.log')

    if dataset.has_projections:
        logger.info("Projections already exist, squashing skipped.")
        sys.exit()

    if not dataset.raw.exists():
        logger.error(
            "Folder raw/ does not exist. Make sure to move all your images under raw/"
        )
        raise FileNotFoundError

    dataset.write_projections()
    logger.info("Projections saved under %s" % dataset.projections)


if __name__ == "__main__":
    args = argparse.Namespace(dataset=Path('data/dataset_test'))
    run(args)



import argparse
from pathlib import Path

import pandas as pd
from stardist.models import StarDist2D

from cenfind.core.data import Dataset, Field
from cenfind.core.measure import field_score
from cenfind.core.detectors import get_model
from cenfind.core.constants import datasets, PREFIX_REMOTE

stardist_model = StarDist2D.from_pretrained('2D_versatile_fluo')


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('model', type=Path)
    parser.add_argument('channel_nuclei', type=int, help='the index of the position')
    args = parser.parse_args()

    return args


def main():
    args = get_args()

    # model_spotnet = get_model(args.model)
    model_stardist = StarDist2D.from_pretrained('2D_versatile_fluo')

    scored = []
    for dataset_name in datasets:
        dataset = Dataset(PREFIX_REMOTE / dataset_name)
        dataset.visualisation.mkdir(exist_ok=True)
        test_files = dataset.pairs('test')

        for field, channel in test_files:
            channel = int(channel)
            score = field_score(field=field, model_nuclei=model_stardist, model_foci=args.model,
                                nuclei_channel=args.channel_nuclei, channel=channel, factor=256,
                                vicinity=-10)
            scored.append(score)

    scores = pd.DataFrame(scored)
    scores.to_csv('out/scores_test.csv')


if __name__ == '__main__':
    main()



import argparse
import os
from pathlib import Path

import pandas as pd
import tifffile as tif
from tqdm import tqdm

from cenfind.core.data import Dataset
from cenfind.core.detectors import extract_foci, extract_nuclei
from cenfind.core.log import get_logger
from cenfind.core.measure import assign, save_foci
from cenfind.core.outline import visualisation


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "score", help="Score the projections given the channels"
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset")
    parser.add_argument("model", type=Path, help="Absolute path to the model folder")
    parser.add_argument(
        "--channel_nuclei",
        "-n",
        type=int,
        required=True,
        help="Channel index for nuclei segmentation, e.g., 0 or 3",
    )
    parser.add_argument(
        "--channel_centrioles",
        "-c",
        nargs="+",
        type=int,
        required=True,
        help="Channel indices to analyse, e.g., 1 2 3",
    )
    parser.add_argument(
        "--vicinity",
        "-v",
        type=int,
        default=50,
        help="Distance threshold in pixel (default: 50 px)",
    )
    parser.add_argument(
        "--factor",
        type=int,
        default=256,
        help="Factor to use: given a 2048x2048 image, 256 if 63x; 2048 if 20x.",
    )
    parser.add_argument("--cpu", action="store_true", help="Only use the cpu")

    return parser


def run(args):
    if not Path(args.model).exists():
        raise FileNotFoundError(f"{args.model} does not exist")

    if args.channel_nuclei in set(args.channel_centrioles):
        raise ValueError("Nuclei channel cannot be in channels")

    if args.cpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    dataset = Dataset(args.dataset)
    logger = get_logger(__name__, file=dataset.logs / 'cenfind.log')

    if not any(dataset.projections.iterdir()):
        logger.error(
            "The projection folder (%s) is empty.\nPlease ensure you have run `squash` or that you have put the projections under projections/"
            % dataset.projections, exc_info=True
        )
        raise FileNotFoundError

    channels = dataset.fields[0].projection.shape[0]

    if args.channel_nuclei not in range(channels):
        logger.error("Index for nuclei (%s) out of index range" % args.channel_nuclei, exc_info=True)
        raise ValueError

    if not set(args.channel_centrioles).issubset(set(range(channels))):
        logger.error(
            "Channels (%s) out of channel range %s" % args.channel_centrioles,
            set(range(channels)),
        )
        raise ValueError

    path_visualisation_model = dataset.visualisation / args.model.name
    path_visualisation_model.mkdir(exist_ok=True)

    pbar = tqdm(dataset.fields)
    from cenfind.core.measure import score

    scores = []
    for field in pbar:
        pbar.set_description(f"{field.name}")
        logger.info("Processing %s" % field.name)

        for channel in args.channel_centrioles:
            nuclei = extract_nuclei(field, args.channel_nuclei, args.factor)
            if len(nuclei) == 0:
                logger.warning("No nucleus has been detected in %s" % field.name)
                continue
            logger.info("Processing %s / %d" % (field.name, channel))
            foci = extract_foci(data=field, foci_model_file=args.model, channel=channel)
            if len(foci) == 0:
                logger.warning("No centrioles (channel: %s) has been detected in %s" % (channel, field.name))

            nuclei_scored = assign(nuclei, foci, vicinity=args.vicinity)
            scored = score(field, nuclei_scored, channel)
            scores.append(scored)

            predictions_path = (
                    dataset.predictions
                    / "centrioles"
                    / f"{field.name}{dataset.projection_suffix}_C{channel}.txt"
            )
            save_foci(foci, predictions_path)

            pbar.set_postfix(
                {
                    "field": field.name,
                    "channel": channel,
                    "nuclei": len(nuclei),
                    "foci": len(foci),
                }
            )

            logger.info(
                "(%s), channel %s: nuclei: %s; foci: %s"
                % (field.name, channel, len(nuclei), len(foci))
            )

            logger.info(
                "Writing visualisations for (%s), channel %s" % (field.name, channel)
            )
            vis = visualisation(
                field, nuclei_scored, foci, channel, args.channel_nuclei
            )
            tif.imwrite(
                path_visualisation_model / f"{field.name}_C{channel}_pred.png", vis
            )

        logger.info("DONE (%s)" % field.name)

    flattened = [leaf for tree in scores for leaf in tree]
    scores_df = pd.DataFrame(flattened)
    scores_df.to_csv(dataset.statistics / "scores_df.tsv", sep="\t", index=False)
    logger.info("Writing raw scores to %s" % str(dataset.statistics / "scores_df.tsv"))


if __name__ == "__main__":
    args = argparse.Namespace(dataset=Path('data/dataset_test'),
                              model=Path('models/master'),
                              channel_nuclei=0,
                              channel_centrioles=[1, 2],
                              vicinity=50,
                              cpu=False,
                              factor=256)
    run(args)



import argparse
from pathlib import Path

import pytomlpp

from cenfind.core.data import Dataset, choose_channel


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "prepare", help="Prepare the dataset structure"
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset")
    parser.add_argument(
        "--projection_suffix",
        type=str,
        default="",
        help="Suffix indicating projection, e.g., `_max` or `Projected`, empty if not specified",
    )
    parser.add_argument(
        "--splits",
        type=int,
        nargs="+",
        help="Write the train and test splits for continuous learning using the channels specified",
    )

    return parser


def run(args):
    """Set up the dataset folder with default directories"""

    with open(args.dataset / "metadata.toml", "w") as f:
        content = pytomlpp.dumps({"projection_suffix": args.projection_suffix})
        f.write(content)

    dataset = Dataset(args.dataset, projection_suffix=args.projection_suffix)

    dataset.setup()
    dataset.write_fields()

    # TODO: Move the rest to training specific programs
    if args.splits:
        train_fields, test_fields = dataset.split_pairs(p=0.9)
        pairs_train = choose_channel(train_fields, args.splits)
        pairs_test = choose_channel(test_fields, args.splits)

        with open(dataset.path / "train.txt", "w") as f:
            for fov, channel in pairs_train:
                f.write(f"{fov.name},{channel}\n")

        with open(dataset.path / "test.txt", "w") as f:
            for fov, channel in pairs_test:
                f.write(f"{fov.name},{channel}\n")


if __name__ == "__main__":
    args = argparse.Namespace(dataset=Path('data/dataset_test'),
                              projection_suffix='_max',
                              splits=[1, 2])
    run(args)



import itertools as it

import numpy as np
from matplotlib import pyplot as plt


def str_to_tuple(position):
    row = ord(position[:1].upper()) - 65
    col = int(position[1:])
    return row, col


def fraction_zero(x):
    return x["0"] / x[list("01234+")].sum()


def prepare_data(data):
    data.columns = ["fov", "channel", "0", "1", "2", "3", "4", "+"]
    data[["well", "field"]] = data["fov"].str.split("_", expand=True)
    data = data.set_index(["well", "field", "channel"])
    data = data.drop(["fov"], axis=1)
    return data


def reduce_data(data):
    summed = data.groupby(["well", "channel"]).sum()
    summed["frac_zero"] = summed.apply(fraction_zero, axis=1)
    summed = summed.reset_index()
    return summed


def plot_layout(data, channel, ax, vmin, vmax):
    vmin = str_to_tuple(vmin)
    vmax = str_to_tuple(vmax)
    _min = data[vmin]
    _max = data[vmax]
    rows, cols = data.shape
    ax.set_title(f"Channel {channel}")
    ax.imshow(data, cmap="cividis", vmin=_min, vmax=_max)
    rows_labels = list("abcdefgh".upper())
    cols_labels = [str(i + 1) for i in range(cols)]
    ax.set_xticks(np.arange(cols), labels=cols_labels)
    ax.set_yticks(np.arange(rows), labels=rows_labels)

    for i, j in it.product(range(rows), range(cols)):
        ax.text(j, i, data[i, j].round(2), ha="center", va="center", color="w")

    return ax


def reshape_data(data, channel, shape):
    rows, cols = shape
    return (
        data.loc[data["channel"] == channel, "frac_zero"].to_numpy().reshape(rows, cols)
    )


def generate_figure(data, vmin, vmax):
    shape = (8, 12)
    channels = data["channel"].unique()
    fig, axes = plt.subplots(nrows=len(channels), ncols=1, figsize=(5, 7))
    for c, channel in enumerate(channels):
        ax = axes[c]
        summed_reshaped = reshape_data(data, channel, shape=shape)
        plot_layout(summed_reshaped, channel, ax, vmin=vmin, vmax=vmax)

    fig.suptitle("Fraction of centriole-free cells")
    fig.tight_layout()
    return fig



import argparse
from pathlib import Path

import tifffile as tf

from cenfind.core.data import Dataset
from cenfind.core.outline import visualisation, Centre


def register_parser(parent_subparsers: argparse.ArgumentParser):
    parser = parent_subparsers.add_parser(
        "predict",
        help="Predict centrioles on new dataset and save under visualisations/runs/<model_name>",
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset folder")
    parser.add_argument("model", type=Path, help="Path to the model")
    parser.add_argument(
        "--channel_nuclei", type=int, required=True, help="Index of nuclei channel"
    )

    return parser


def run(args):
    dataset = Dataset(args.dataset)

    model_predictions = dataset.visualisation / "runs"
    model_predictions.mkdir(exist_ok=True)

    model_run = model_predictions / args.model.name
    model_run.mkdir(exist_ok=True)

    from cenfind.core.measure import extract_foci

    for field, channel in dataset.pairs("test"):
        foci = extract_foci(field, args.model, channel, prob_threshold=0.5)
        foci = [Centre((r, c), f_id, "Centriole") for f_id, (r, c) in enumerate(foci)]
        print(
            "Writing visualisations for field: %s, channel: %s, %s foci detected"
            % (field.name, channel, len(foci))
        )
        vis = visualisation(
            field, foci, channel, channel_nuclei=args.channel_nuclei
        )
        tf.imwrite(model_run / f"{field.name}_C{channel}_pred.png", vis)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cenfind',
 'cenfind.cli',
 'cenfind.core',
 'cenfind.experiments',
 'cenfind.labelbox',
 'cenfind.publication']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=1.3.0,<2.0.0',
 'csbdeep>=0.7.3,<0.8.0',
 'labelbox>=3.36.1,<4.0.0',
 'llvmlite==0.39.1',
 'numba==0.56.4',
 'numpy>=1.23.5,<2.0.0',
 'opencv-python>=4.7.0.72,<5.0.0.0',
 'ortools==9.4.1874',
 'pandas>=1.4.1,<2.0.0',
 'protobuf==3.19.6',
 'python-dotenv>=0.21.1,<0.22.0',
 'pytomlpp>=1.0.10,<2.0.0',
 'scikit-image>=0.19.2,<0.20.0',
 'scikit-learn>=1.2.1,<2.0.0',
 'scipy>=1.7.1,<2.0.0',
 'spotipy-detector>=0.1.0,<0.2.0',
 'stardist>=0.8.3,<0.9.0',
 'tifffile>=2022.5.4,<2023.0.0',
 'tqdm>=4.62.3,<5.0.0']

extras_require = \
{':sys_platform == "arm64"': ['tensorflow-macos==2.9.0'],
 ':sys_platform == "win32" or sys_platform == "linux"': ['tensorflow>=2.9.0,<3.0.0']}

entry_points = \
{'console_scripts': ['cenfind = cenfind.__main__:main',
                     'evaluate = cenfind.cli.evaluate:main',
                     'upload = cenfind.labelbox.upload_labels:main',
                     'vignettes = cenfind.labelbox.vignettes:main']}

setup_kwargs = {
    'name': 'cenfind',
    'version': '0.11.5',
    'description': 'Score cells for centrioles in IF data',
    'long_description': "# CenFind\n\nA command line interface to score cells for centrioles.\n\n## Introduction\n\n`cenfind` is a command line interface to detect and assign centrioles in immunofluorescence images of human cells. Specifically, it orchestrates:\n\n- the z-max projection of the raw files;\n- the detection of centrioles;\n- the detection of the nuclei;\n- the assignment of the centrioles to the nearest nucleus.\n\n## Installation\n1. Install python via pyenv\n2. Download and set up 3.9.5 as local version\n3. Set up Python interpreter\n```shell\npyenv local 3.9.5\npyenv global 3.9.5\n```\n4. Create a virtual environment for CenFind\n```shell\npython -m venv venv-cenfind\nsource venv-cenfind/bin/activate\n```\n\n5. Check that `cenfind`'s programs are correctly installed by running:\n\n```shell\ncenfind squash --help\n```\n\n## Basic usage\nBefore scoring the cells, you need to prepare the dataset folder. \n`cenfind` assumes a fixed folder structure. \nIn the following we will assume that the .ome.tif files are all immediately in raw/. \nEach file is a z-stack field of view (referred to as field, in the following) containing 4 channels (0, 1, 2, 3). The channel 0 contains the nuclei and the channels 1-3 contains centriolar markers.\n\n```text\n<project_name>/\n└── raw/\n```\n2. Run `prepare` to initialise the dataset folder with a list of output folders:\n```shell\ncenfind prepare /path/to/dataset --splits 1 2 --projection_suffix _max\n```\n```shell\nusage: CENFIND prepare [-h] [--projection_suffix PROJECTION_SUFFIX] [--splits SPLITS [SPLITS ...]] dataset\n\npositional arguments:\n  dataset               Path to the dataset\n\noptions:\n  -h, --help            show this help message and exit\n  --projection_suffix PROJECTION_SUFFIX\n                        Suffix indicating projection, e.g., `_max` or `Projected`, empty if not specified (default: )\n  --splits SPLITS [SPLITS ...]\n                        Write the train and test splits for continuous learning using the channels specified (default: None)\n```\n\n2. Run `squash` with the path to the project folder and the suffix of the raw files. `projections/` is populated with the max-projections `*_max.tif` files.\n```shell\ncenfind squash path/to/dataset\n```\n```shell\nusage: CENFIND squash [-h] dataset\n\npositional arguments:\n  dataset     Path to the dataset folder\n\noptions:\n  -h, --help  show this help message and exit\n\n```\n\n3. Run `score` with the arguments source, the index of the nuclei channel (usually 0 or 3), the channel to score and the path to the model. You need to download it from https://figshare.com/articles/software/Cenfind_model_weights/21724421\n```shell\ncenfind score /path/to/dataset /path/to/model/ -n 0 -c 1 2 3 -v 50\n```\n```shell\nusage: CENFIND score [-h] --channel_nuclei CHANNEL_NUCLEI --channel_centrioles CHANNEL_CENTRIOLES [CHANNEL_CENTRIOLES ...] [--vicinity VICINITY] [--factor FACTOR] [--cpu]\n                     dataset model\n\npositional arguments:\n  dataset               Path to the dataset\n  model                 Absolute path to the model folder\n\noptions:\n  -h, --help            show this help message and exit\n  --channel_nuclei CHANNEL_NUCLEI, -n CHANNEL_NUCLEI\n                        Channel index for nuclei segmentation, e.g., 0 or 3 (default: None)\n  --channel_centrioles CHANNEL_CENTRIOLES [CHANNEL_CENTRIOLES ...], -c CHANNEL_CENTRIOLES [CHANNEL_CENTRIOLES ...]\n                        Channel indices to analyse, e.g., 1 2 3 (default: None)\n  --vicinity VICINITY, -v VICINITY\n                        Distance threshold in pixel (default: 50 px) (default: 50)\n  --factor FACTOR       Factor to use: given a 2048x2048 image, 256 if 63x; 2048 if 20x: (default: 256)\n  --cpu                 Only use the cpu (default: False)\n```\n\n4. Check that the predictions are satisfactory by looking at the folders `visualisations/` and `statistics/`\n\n5. If you are interested in categorising the number of centrioles, run `cenfind analyse path/to/dataset --by <well>` the --by option is interesting if you want to group your scoring by well, if the file names obey to the rule `<WELLID_FOVID>`.\n\n```shell\nusage: CENFIND analyse [-h] --by BY dataset\n\npositional arguments:\n  dataset     Path to the dataset\n\noptions:\n  -h, --help  show this help message and exit\n  --by BY     Grouping (field or well) (default: None)\n```\n\n## Running `cenfind score` in the background\n\nWhen you exit the shell, running programs receive the SIGHUP, which aborts them. This is undesirable if you need to close your shell for some reasons. Fortunately, you can make your program ignore this signal by prepending the program with the `nohup` command. Moreover, if you want to run your program in the background, you can append the ampersand `&`. In practice, run `nohup cenfind score ... &` instead of `cenfind score ...`.\n\nThe output will be written to the file `nohup.out` and you can peek the progress by running `tail -F nohup.out`, the flag `-F` will refresh the screen as the file is being written. Enter Ctrl-C to exit the tail program.\n\nIf you want to kill the program score, run  `jobs` and then run `kill <jobid>`. If you see no jobs, check the log `nohup.out`; it can be done or the program may have crashed, and you can check the error there.\n\n## Evaluating the quality of the model on a new dataset\n\nThe initial model M is fitted using a set of five representative datasets, hereafter referred to as the standard datasets (DS1-5). \nIf your type of data deviates too much from the standard dataset, M may perform less well. \n\nSpecifically, when setting out to score a new dataset, you may be faced with one of three situations, as reflected by the corresponding F1 score (i.e., 2TP/2TP+FN+FP, TP: true positive, FP: false positive; FN: false negative): \n(1) the initial model (M) performs well on the new dataset (0.9 ≤ F1 ≤ 1); in this case, model M is used; \n(2) model M performs significantly worse on the new dataset (0.5 ≤ F1 < 0.9); in this case, you may want to consider retraining the model (see below); \n(3) the model does not work at all (0 ≤  F1 < 0.5); such a low F1value probably means that the features of the data set are too distant from the original representative data set to warrant retraining starting from M. \n\nBefore retraining a model (2), verify once more the quality of the data, which needs to be sufficiently good in terms of signal over noise to enable efficient learning. \nIf this is not the case, it is evident that the model will not be able to learn well. \nIf you, as a human being, cannot tell the difference between a real focus and a stray spot using a single channel at hand (i.e., not looking at other channels), the same will hold for the model. \n\nTo retrain the model, you first must annotate the dataset, divide it randomly into training and test sets (90 % versus 10 % of the data, respectively). \nNext, the model is trained with the 90 % set, thus generating a new model, M*. \nLast, you will evaluate the gain of performance on the new dataset, as well as the potential loss of performance on the standard datasets. \n\n### Detailed training procedure:\n1.\tSplit the dataset into training (90%) and test (10%) sets, each containing one field of view and the channel to use. This helps trace back issues during the training and renders the model fitting reproducible.\n```shell\n```\n2.\tLabel all the images present in training and test sets using Labelbox. To upload the images, please create the vignettes first and then upload them once you have a project set up.\n```shell\ncenfind vignettes /path/to/dataset\ncenfind upload /path/to/dataset --env /path/to/.env\n```\n3.\tSave all foci coordinates (x, y), origin at top-left, present in one field of view as one text file under /path/to/dataset/annotation/centrioles/ with the naming scheme <dataset_name>_max_C<channel_index>.txt.\n```shell\ncenfind download dataset-name --env /path/to/.env\n```\n4.\tEvaluate the newly annotated dataset using the model M by computing the F1 score.\nevaluate dataset model\n\n```shell\nusage: CENFIND evaluate [-h] [--performances_file PERFORMANCES_FILE] [--tolerance TOLERANCE] --channel_nuclei CHANNEL_NUCLEI --channel_centrioles CHANNEL_CENTRIOLES [CHANNEL_CENTRIOLES ...]\n                        [--vicinity VICINITY]\n                        dataset model\n\npositional arguments:\n  dataset               Path to the dataset folder\n  model                 Path to the model\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --performances_file PERFORMANCES_FILE\n                        Path of the performance file, STDOUT if not specified (default: None)\n  --tolerance TOLERANCE\n                        Distance in pixels below which two points are deemed matching (default: 3)\n  --channel_nuclei CHANNEL_NUCLEI\n                        Channel index for nuclei segmentation, e.g., 0 or 3 (default: None)\n  --channel_centrioles CHANNEL_CENTRIOLES [CHANNEL_CENTRIOLES ...]\n                        Channel indices to analyse, e.g., 1 2 3 (default: None)\n  --vicinity VICINITY   Distance threshold in micrometer (default: -5 um) (default: -5)\n```\n5.\tIf the performance is poor (i.e., F1 score < 0.9), fit a new model instance, M*, with the standard dataset plus the new dataset (90% in each case).\n6.\tTest performance of model M* on the new data set; hopefully the F1 score will now be ≥ 0.9 (if not: consider increasing size of annotated data).\n7.\tTest performance of model M* on the standard datasets; if performance of F1* ≥ F1, then save M* as the new M (otherwise keep M* as a separate model for the new type of data set).\n\n\n## Internal API\n\n`cenfind` consists of two core classes: `Dataset` and `Field`.\n\nsetup function that \n- write the toml file\n- write the field.txt that list all the files\n- create the folders projections, predictions, visualisations and statistics\n\nA `Dataset` represents a collection of related fields and should:\n- construct file names for fields\n- split the fileds into train, test and validate\n- load all fields or only for a specific split\n\nA `Field` represents a field of view and should:\n- get Dataset\n- load the projection as np.ndarray\n- load the channel as np.ndarray\n- detect centrioles => list of Points\n- detect nuclei => list of Contours\nif present:\n- load annotation as np.ndarray\n- load mask as np.ndarray\n\nA scoring function that \n- assigns centrioles to nuclei (contours, points) => pairs\n- compares predictions with annotation (points, points) => metrics_namespace\n\nVisualisation functions that:\n- outline centrioles and nuclei (data, points) => image\n- create composite vignettes (data) => composite_image\n- flag partial nuclei (contours, tolerance) => contours\n",
    'author': 'Leo Burgy',
    'author_email': 'leo.burgy@epfl.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/UPGON/cenfind',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
